from docx.oxml import register_element_cls, CT_R
from docx.text.paragraph import Paragraph
from docx.oxml.ns import qn
from docx.text.run import Run

from pymdocx.oxml.revision import CT_Rev, CT_Deltext
from pymdocx.text.revision import Revision

register_element_cls('w:ins', CT_Rev)
register_element_cls('w:del', CT_Rev)
register_element_cls('w:delText', CT_Deltext)


def monkey_patch_run():
    def del_text(self):
        del_text_list = self._element.findall(qn('w:delText'))
        text = ''
        for dt in del_text_list:
            text += dt.text
        return text
    property_ = property(del_text, None, None)
    setattr(Run, 'del_text', property_)


monkey_patch_run()


def monkey_patch_paragraph():
    def revisions(self):
        rev_ins = self._element.findall(qn('w:ins'))
        rev_del = self._element.findall(qn('w:del'))
        return [Revision(r, self) for r in rev_ins + rev_del]
    setattr(Paragraph, 'revisions', property(revisions, None, None))

    def all_runs_no_ins(self):
        return [Run(r, self) for r in self._p.xpath('.//w:r[not(ancestor::w:r) and not(ancestor::w:ins)]')]
    setattr(Paragraph, 'all_runs_no_ins', property(all_runs_no_ins, None, None))

    def origin_text(self):
        return ''.join([r.del_text if r.del_text else r.text for r in self.all_runs_no_ins])
    setattr(Paragraph, 'origin_text', property(origin_text, None, None))


monkey_patch_paragraph()



import os

from pymdocx.common.utils import print_doc, get_element_comment_revision_matrix, get_doc

DIR_PATH = './../data/other'


def test_print_doc():
    doc_path = os.path.join(DIR_PATH, 'pt1.docx')
    print_doc(doc_path)


def test_get_element_comment_revision_matrix():
    DIR_PATH = './../data/test_p'
    doc_path_o = os.path.join(DIR_PATH, 'pt0_0.docx')
    doc_path_a = os.path.join(DIR_PATH, 'pt0_a.docx')
    doc_o = get_doc(doc_path_o)
    doc_a = get_doc(doc_path_a)
    print('\n')
    print(get_element_comment_revision_matrix(doc_a))


"""
Custom element classes related to the revisions part
"""
from docx.oxml.simpletypes import ST_DecimalNumber, ST_String
from docx.oxml.xmlchemy import BaseOxmlElement, RequiredAttribute, ZeroOrMore


class CT_Rev(BaseOxmlElement):
    """
    ``<w:ins>`` and ``<w:del>`` element, a container for Comment properties
    """
    _id = RequiredAttribute('w:id', ST_DecimalNumber)
    date = RequiredAttribute('w:date', ST_String)
    author = RequiredAttribute('w:author', ST_String)

    r = ZeroOrMore('w:r')


class CT_Deltext(BaseOxmlElement):
    """
    ``<w:del>`` element, a container for Comment properties
    """


import os

from pymdocx.common.comment import get_comment_xml

DIR_PATH = './../data/test_p'


def test_get_comment_xml():
    doc_path = os.path.join(DIR_PATH, 'pt0_a.docx')
    comment_xml = get_comment_xml(doc_path)
    print(comment_xml)


import docx
from pymdocx.common.constants import WORD_NAMESPACE


def create_comment_doc(doc_file_path, doc_file_path_new):
    # 创建一个有评论的doc
    document = docx.Document(docx=doc_file_path)
    # for p in document.paragraphs:
    #     print(p)
    #     comment = p.add_comment('测试', author='Obay Daba', initials='od')
    document.save(doc_file_path_new)


def has_comment(p):
    return 1 if p.comments else 0


def get_comment(doc_file_path):
    document = docx.Document(docx=doc_file_path)
    for p in document.paragraphs:
        for c in p.comments:
            print(c.text)


def get_comment_xml(doc_file_path):
    document = docx.Document(docx=doc_file_path)
    comment_xml = document.comments_part.element.xml
    return comment_xml


def parse_p_comment(p_obj):
    # docx.text.comment.Comment
    comment_info_list = []
    for comment in p_obj.comments:
        comment_info_list.append({
            'comment_text': comment.text,
            'dtime': comment.element.date,
            'author': comment.element.author,
            'initials': comment.element.initials,
            'id': comment.element._id,
        })
    return comment_info_list


def add_new_comment(comment_part, author, dtime, comment_text, initials=''):
    comment = comment_part.add_comment(author, initials, dtime)
    comment._add_p(comment_text)
    return comment._id


def add_p_comment_next(p1, p2, comments_part_obj):
    # 添加段落p及其批注
    p1._p.addnext(p2._p)
    comment_list = parse_p_comment(p2)
    new_comment_id_list = []
    for comment in comment_list:
        comment_id = add_new_comment(comments_part_obj, comment['author'], comment['dtime'],
                                     comment['comment_text'],
                                     initials=comment['initials'])
        new_comment_id_list.append(comment_id)
    modify_oxml_element_comment_id(p2._p, new_comment_id_list)


def modify_oxml_element_comment_id(oxml_element_obj, new_comment_id_list):
    attr_name = '{' + WORD_NAMESPACE['w'] + '}' + 'id'
    for i, cs in enumerate(oxml_element_obj.xpath('.//w:commentRangeStart')):
        cs.attrib[attr_name] = str(new_comment_id_list[i])
    for i, ce in enumerate(oxml_element_obj.xpath('.//w:commentRangeEnd')):
        ce.attrib[attr_name] = str(new_comment_id_list[i])
    for i, cr in enumerate(oxml_element_obj.xpath('.//w:commentReference')):
        cr.attrib[attr_name] = str(new_comment_id_list[i])


def add_comment_2_p_end(p1, p2, comment_part_obj):
    comment_list = parse_p_comment(p2)
    for c in comment_list:
        p1.add_comment(c['comment_text'],
                       author=c['author'],
                       dtime=c['dtime'],
                       comment_part=comment_part_obj,
                       rangeStart=len(p1._p.getchildren()),
                       rangeEnd=len(p1._p.getchildren()))


import os
from pprint import pprint

from pymdocx.common.utils import get_doc
from pymdocx.doc.paragraph import merge_paragraph_comment_revision, parse_paragraph_differences

DIR_PATH = './../data/test_p'
OUTPUT_PATH = './../data/output'


def test_parse_paragraph_differences():

    doc_file_path_o = os.path.join(DIR_PATH, "pt0_0.docx")
    doc_file_path_a = os.path.join(DIR_PATH, "pt0_a.docx")
    doc_file_path_c = os.path.join(DIR_PATH, "pt0_c.docx")

    doc_o = get_doc(doc_file_path_o)
    doc_a = get_doc(doc_file_path_a)
    doc_c = get_doc(doc_file_path_c)
    print(1)
    pprint(parse_paragraph_differences(doc_o, [doc_a, doc_c]))


def test_merge_paragraph_comment_revision():

    doc_file_path_o = os.path.join(DIR_PATH, "pt0_0.docx")
    doc_file_path_a = os.path.join(DIR_PATH, "pt0_a.docx")
    doc_file_path_c = os.path.join(DIR_PATH, "pt0_c.docx")
    output_file_path = os.path.join(OUTPUT_PATH, "pt0_new_v2.docx")

    doc_o = get_doc(doc_file_path_o)
    doc_a = get_doc(doc_file_path_a)
    doc_c = get_doc(doc_file_path_c)

    merge_paragraph_comment_revision(doc_o, [doc_a, doc_c])
    doc_o.save(output_file_path)


def test_p_style():
    doc_file_path_c_1 = os.path.join(DIR_PATH, "pt0_c_1.docx")
    doc_c_1 = get_doc(doc_file_path_c_1)
    print(1)
    for r in doc_c_1.paragraphs[2].runs:
        r.bold = True
        r.italic = True


from distutils.core import setup
from setuptools import find_packages


with open("README.rst", "r") as f:
  long_description = f.read()


setup(name='pymdocx',  # 包名
      version='0.3.9',  # 版本号
      description='pydocx is a Python library for updating Microsoft Word files.',
      long_description=long_description,
      author='mirrornight',
      author_email='mirrornighth@gmail.com',
      url='https://github.com/mirrornight/pydocx',
      install_requires=['bayoo-docx==0.2.17'],
      license='MIT License',
      packages=find_packages(),
      platforms=["all"],
      classifiers=[
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
      ],
      )


import os

from pymdocx.common.utils import get_doc
from pymdocx.common.revision import get_ins_in_paragraph

DIR_PATH = './../data/test_p'


def test_get_ins_in_paragraph():
    doc_path = os.path.join(DIR_PATH, 'pt0_a.docx')
    doc_obj = get_doc(doc_path)
    tp = doc_obj.paragraphs[1]
    ins_xml = get_ins_in_paragraph(tp)
    print(ins_xml)


from pymdocx.common.comment import has_comment, add_p_comment_next, add_comment_2_p_end
from pymdocx.common.revision import has_revision, add_revision_2_p_end, remove_revision
from pymdocx.common.utils import _get_actual_p_index
from pymdocx.doc.paragraph import _merge_p, p_bold_italic


def detect_comment_revision_in_table(doc_obj_list):
    """
    {
    table_index_1: {cell_index: {p_index: [doc_index_list]}}
    }
    """
    m = {}
    for doc_index, doc_obj in enumerate(doc_obj_list):
        for t_index, t in enumerate(doc_obj.tables):
            for c_index, c in enumerate(t._cells):
                for p_index, p in enumerate(c.paragraphs):
                    if has_comment(p) or has_revision(p):
                        if t_index not in m:
                            m[t_index] = {c_index: {p_index: [doc_index]}}
                        elif c_index in m[t_index] and p_index not in m[t_index][c_index]:
                            m[t_index][c_index][p_index] = [doc_index]
                        elif c_index in m[t_index] and p_index in m[t_index][c_index]:
                            m[t_index][c_index][p_index].append(doc_index)
                        elif c_index not in m[t_index]:
                            m[t_index][c_index] = {p_index: [doc_index]}
    return m


def merge_table_comment_revision_stack(doc_base_obj, merge_doc_list):
    detect_dict = detect_comment_revision_in_table(merge_doc_list)
    for t_index, t_v in detect_dict.items():
        for c_index, c_v in t_v.items():
            has_add_p_count = 0
            remove_p_list = []
            has_add_mapping = {}
            for p_index, doc_index_list in c_v.items():
                base_p = last_p = doc_base_obj.tables[t_index]._cells[c_index].paragraphs[p_index + has_add_p_count]
                remove_p_list.append(last_p)
                for i, doc_index in enumerate(doc_index_list):
                    merge_doc_obj = merge_doc_list[doc_index]
                    if merge_doc_obj.tables[t_index]._cells[c_index].paragraphs:
                        actual_p_index = _get_actual_p_index(has_add_mapping, doc_index, p_index)
                        target_p = merge_doc_obj.tables[t_index]._cells[c_index].paragraphs[actual_p_index]
                        add_p_comment_next(last_p, target_p, doc_base_obj.comments_part.element)
                        last_p = target_p
                        if i > 0:
                            p_bold_italic(last_p)
                        has_add_p_count += 1
                    else:
                        if base_p in remove_p_list:
                            remove_p_list.remove(base_p)
            [rp.delete() for rp in remove_p_list]


def merge_table_comment_revision_end(doc_base_obj, merge_doc_list):
    detect_dict = detect_comment_revision_in_table(merge_doc_list)
    for t_index, t_v in detect_dict.items():
        for c_index, c_v in t_v.items():
            has_add_p_count = 0
            remove_p_list = []
            has_add_mapping = {}
            for p_index, doc_index_list in c_v.items():
                base_p = last_p = doc_base_obj.tables[t_index]._cells[c_index].paragraphs[p_index + has_add_p_count]
                remove_p_list.append(last_p)
                for i, doc_index in enumerate(doc_index_list):
                    merge_doc_paragraphs = merge_doc_list[doc_index].tables[t_index]._cells[c_index].paragraphs
                    if merge_doc_paragraphs:
                        has_add_p_count, last_p = _merge_p(i, has_add_mapping, doc_index, p_index, merge_doc_paragraphs, last_p,
                                               doc_base_obj.comments_part.element, has_add_p_count, remove_p_list)
                    else:
                        if base_p in remove_p_list:
                            remove_p_list.remove(base_p)
            [rp.delete() for rp in remove_p_list]


merge_table_comment_revision = merge_table_comment_revision_stack